from typing import List, Optional

import numpy as np  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
from matplotlib import patches

from .bounding_box import BoundingBox


class Taxonomy:
    """Classes for detection."""


class DataPoint:
    def __init__(
        self, org_id: str, label_set_name: str, dp_id: str, image: np.ndarray
    ) -> None:
        """Construct DataPoint."""
        self.org_id = org_id
        self.label_set_name = label_set_name
        self.dp_id = dp_id
        self.image: np.ndarray = image

        self.gt_boxes: Optional[List[BoundingBox]] = None
        self.gt_boxes_classes: Optional[List[str]] = None

    def __repr__(self) -> str:
        """Get a str representation of DataPoint."""
        return f"{self.__class__.__name__}<dp={self.dp_id}>"

    @property
    def height(self) -> int:
        return self.image.shape[0]  # type: ignore

    @property
    def width(self) -> int:
        return self.image.shape[1]  # type: ignore

    def show_image(self, show_gt: bool = True, show_pred: bool = False) -> None:
        """Use matplotlib to show the image."""
        colors = ["xkcd:blue", "xkcd:red", "xkcd:green", "xkcd:orange", "xkcd:purple"]

        fig, ax = plt.subplots(1)
        if show_gt and self.gt_boxes and self.gt_boxes_classes:
            for ii, box in enumerate(self.gt_boxes):
                object_ = box.as_array()
                color = colors[ii % (len(colors) - 1)]
                height = object_[3] * self.height
                width = object_[2] * self.width

                bottom_left = (object_[0] * self.width, object_[1] * self.height)
                rect = patches.Rectangle(
                    bottom_left,
                    width,
                    height,
                    linewidth=1.5,
                    edgecolor=color,
                    facecolor="none",
                )
                ax.add_patch(rect)
                ax.text(
                    bottom_left[0] + 1.4,
                    bottom_left[1] - 2,
                    str(self.gt_boxes_classes[ii]),
                    backgroundcolor=color,
                    fontsize=10,
                )

        ax.imshow(self.image)
        fig.tight_layout()
        plt.show()



"""Getting data from redbrick api."""

import redbrick
import json
from dataclasses import dataclass, asdict
from typing import Dict, List, Any, Optional
import requests
import numpy as np  # type: ignore
import cv2  # type: ignore
from .api_base import RedBrickApiBase
from .entity import DataPoint, BoundingBox


class RedBrickApi(RedBrickApiBase):
    """Implement Abstract API."""

    def __init__(self, cache: bool = False, custom_url: Optional[str] = None) -> None:
        """Construct RedBrickApi."""
        self.cache: Dict[str, Dict[str, Dict[str, DataPoint]]] = {}
        self.client = redbrick.client.RedBrickClient()
        if custom_url:
            self.url = custom_url
        else:
            self.url = "https://redbrick-prod-1.herokuapp.com/graphql/"

    def get_datapoint_ids(self, org_id: str, label_set_name: str) -> List[str]:
        """Get a list of datapoint ids in labelset."""
        query_string = """
            query ($orgId:UUID!, $name:String!) {
                customGroup(orgId: $orgId, name: $name) {
                    datapoints(skip:0, first:-1) {
                        dpId
                    }
                }
            }
        """

        query_variables = {"orgId": org_id, "name": label_set_name}
        query = dict(query=query_string, variables=query_variables)

        result = self._execute_query(query)

        all_dp_ids = [dp["dpId"] for dp in result["customGroup"]["datapoints"]]
        return all_dp_ids

    def get_datapoint(self, org_id: str, label_set_name: str, dp_id: str) -> DataPoint:
        """Get all relevant information related to a datapoint."""
        temp = self.cache.get(org_id, {}).get(label_set_name, {}).get(dp_id)
        if temp:
            return temp

        query_string = """
            query ($orgId: UUID!, $dpId: UUID!, $name:String!) {
                labelData(orgId: $orgId, dpId: $dpId, customGroupName: $name) {
                    blob
                    dataPoint {
                        items(presigned:true)
                    }
                }
            }
        """
        query_variables = {"orgId": org_id, "name": label_set_name, "dpId": dp_id}
        query = dict(query=query_string, variables=query_variables)
        result = self._execute_query(query)
        # parse result
        image_url = result["labelData"]["dataPoint"]["items"][0]
        labels = json.loads(result["labelData"]["blob"])["items"][0]["labels"]

        # get image array
        image = self._url_to_image(image_url)

        dpoint = DataPoint(org_id, label_set_name, dp_id, image)
        # convert labels and initialize ground truth
        dpoint.gt_boxes = [BoundingBox.from_remote(label["bbox2d"]) for label in labels]
        dpoint.gt_boxes_classes = [label["category"][0][0] for label in labels]

        self.cache[org_id] = self.cache.get(org_id, {})
        self.cache[org_id][label_set_name] = self.cache[org_id].get(label_set_name, {})
        self.cache[org_id][label_set_name][dp_id] = dpoint

        return dpoint

    @staticmethod
    def _url_to_image(url: str) -> np.ndarray:
        """Get a cv2 image object from a url."""
        # Download the image, convert it to a NumPy array, and then read
        # it into OpenCV format
        resp = requests.get(url, stream=True)
        resp.raw.decode_content = True
        image = np.asarray(bytearray(resp.raw.read()), dtype="uint8")
        image = cv2.imdecode(image, cv2.IMREAD_COLOR)
        # cv2 returns a BGR image, need to convert to RGB
        # the copy operation makes the memory contiguous for tensorify-ing
        return np.flip(image, axis=2).copy()

    def _execute_query(self, query: Dict) -> Any:
        """Execute a graphql query."""
        headers = {"ApiKey": self.client.api_key}
        try:
            response = requests.post(self.url, headers=headers, json=query)
            return response.json()["data"]
        except ValueError:
            print(response.content)
            print(response.status_code)


# @dataclass
# class GraphQLQuery:
#     """Query to execute on GraphQL."""

#     query: str
#     variables: Dict


# @dataclass
# class DataPoint:
#     """A Datapoint returned by GraphQL query."""

#     image_url: str
#     labels: str


# def get_datapoint_ids(org_id: str, label_set_name: str) -> List[str]:
#     """Get all data points in the label set."""


# def get_datapoint(org_id: str, label_set_name: str, dp_id: str) -> DataPoint:
#     """Get data needed for a specific data point."""
#     query_string = """
#         query ($orgId: UUID!, $dpId: UUID!, $name:String!) {
#             labelData(orgId: $orgId, dpId: $dpId, customGroupName: $name) {
#                 blob
#                 dataPoint {
#                     items(presigned:true)
#                 }
#             }
#         }
#     """
#     query_variables = {"orgId": org_id, "name": label_set_name, "dpId": dp_id}
#     query = GraphQLQuery(query_string, query_variables)
#     result = execute_query(query)
#     return DataPoint(
#         labels=result["labelData"]["blob"],
#         image_url=result["labelData"]["dataPoint"]["items"][0],
#     )



import numpy as np
from typing import Dict


class BoundingBox:
    """."""

    def __init__(self) -> None:
        """Construct BoundingBox."""
        self._xnorm: float
        self._ynorm: float
        self._wnorm: float
        self._hnorm: float

    def __repr__(self) -> str:
        """Get a string representation of object."""
        return f"BBOX<xmin={self._xnorm}, ymin={self._ynorm}, width={self._wnorm}, height={self._hnorm}>"

    def as_array(self, min_max: bool = False) -> np.ndarray:
        """Get array representation of bounding box."""
        if not min_max:
            return np.array([self._xnorm, self._ynorm, self._wnorm, self._hnorm])
        return np.array(
            [
                self._xnorm,
                self._ynorm,
                self._xnorm + self._wnorm,
                self._ynorm + self._hnorm,
            ]
        )

    @classmethod
    def from_remote(cls, obj: Dict) -> "BoundingBox":
        """Convert return value from server into BoundingBox entity."""
        this = cls()
        this._xnorm = obj["xnorm"]
        this._ynorm = obj["ynorm"]
        this._wnorm = obj["wnorm"]
        this._hnorm = obj["hnorm"]
        return this

    @classmethod
    def from_array(cls, obj: np.ndarray) -> "BoundingBox":
        """Convert a numpy array to BoundingBox."""
        if not obj.shape == (1, 4):
            raise ValueError("box.shape must be (1,4)")
        for val in obj:
            if val > 1 or val < 0:
                raise ValueError("Box points must be [0,1]")
        this = cls()
        this._xnorm = obj[0]
        this._ynorm = obj[1]
        this._wnorm = obj[2]
        this._hnorm = obj[3]
        return this

    @staticmethod
    def normalize(boxes: np.ndarray, height: int, width: int) -> np.ndarray:
        """
        Normalize all values of box.

        boxes: nx4
        height: height of image in pixels
        width: width of image in pixels
        """
        boxes[:, 0] = boxes[:, 0] / width
        boxes[:, 2] = boxes[:, 2] / width
        boxes[:, 1] = boxes[:, 1] / height
        boxes[:, 3] = boxes[:, 3] / height
        return boxes

    @staticmethod
    def get_coords_absolute(boxes: np.ndarray, height: int, width: int) -> np.ndarray:
        """
        Normalize all values of box.

        boxes: nx4
        height: height of image in pixels
        width: width of image in pixels
        """
        boxes[:, 0] = width * boxes[:, 0]
        boxes[:, 2] = width * boxes[:, 2]
        boxes[:, 1] = height * boxes[:, 1]
        boxes[:, 3] = height * boxes[:, 3]
        return boxes



"""Abstract interface to RedBrick API."""

from typing import List
from abc import ABC, abstractmethod
from redbrick.entity import DataPoint


class RedBrickApiBase(ABC):
    """Abstract interface to RedBrick API."""

    @abstractmethod
    def get_datapoint_ids(self, org_id: str, label_set_name: str) -> List[str]:
        """Get a list of datapoint ids in labelset."""

    @abstractmethod
    def get_datapoint(self, org_id: str, label_set_name: str, dp_id: str) -> DataPoint:
        """Get all relevant information related to a datapoint."""



"""A higher level abstraction."""

from typing import Optional, List
from random import randint

import numpy as np  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
from matplotlib import patches

from redbrick.api import RedBrickApi
from redbrick.entity import DataPoint


class LabelsetLoader:
    """A basic high level loader class."""

    def __init__(
        self, org_id: str, label_set_name: str, dp_ids: Optional[List[str]] = None
    ) -> None:
        """Construct Loader."""
        self.org_id = org_id
        self.label_set_name = label_set_name
        self.api_client = RedBrickApi(cache=False)

        print("Counting available data points... ", end="")
        if dp_ids:
            self.dp_ids = dp_ids
        else:
            self.dp_ids = self.api_client.get_datapoint_ids(
                self.org_id, self.label_set_name
            )
        print(len(self.dp_ids))

    def __getitem__(self, index: int) -> DataPoint:
        """Get information needed for a single item."""
        return self.api_client.get_datapoint(
            self.org_id, self.label_set_name, self.dp_ids[index]
        )

    def number_of_datapoints(self) -> int:
        """Get number of datapoints."""
        return len(self.dp_ids)

    def show_random_image(self) -> None:
        """Show a random image."""
        idx = randint(0, self.number_of_datapoints() - 1)
        self[idx].show_image(show_gt=True)



"""
Initialize RedBrick package.

Derek Lukacs, 2020
"""

from .client import RedBrickClient
import redbrick.api

# import redbrick.base
import redbrick.client
import redbrick.labelset


def init(api_key: str) -> None:
    """Initialize package state."""
    RedBrickClient(api_key)



"""Utilities for training with RedBrick labelsets."""
import random
from typing import Tuple
import redbrick
from .loader import LabelsetLoader


def training(
    org_id: str, label_set_name: str, validation_split: float = 0.0
) -> Tuple[LabelsetLoader, LabelsetLoader]:
    """Create a training split and give 2 loaders."""
    if validation_split < 0 or validation_split > 1:
        raise Exception("value of `test_split` must be in range [0, 1]")
    dp_ids = redbrick.api.RedBrickApi().get_datapoint_ids(org_id, label_set_name)

    random.shuffle(dp_ids)

    cut_off = int((1 - validation_split) * len(dp_ids))

    train = LabelsetLoader(org_id, label_set_name, dp_ids=dp_ids[:cut_off])
    validate = LabelsetLoader(org_id, label_set_name, dp_ids=dp_ids[cut_off:])

    return train, validate
